from datetime import timedelta
from os import PathLike
from pathlib import Path

from geojson import Point
from shapely.geometry import LineString

from .tracks import APRSTrack

KML_STANDARD = '{http://www.opengis.net/kml/2.2}'


def write_aprs_packet_tracks(packet_tracks: [APRSTrack], output_filename: PathLike):
    if not isinstance(output_filename, Path):
        output_filename = Path(output_filename)
    output_filename = output_filename.resolve().expanduser()
    extension = output_filename.suffix
    if extension == '.geojson':
        import geojson

        features = []
        for packet_track in packet_tracks:
            features.extend(geojson.Feature(geometry=geojson.Point(packet.coordinates.tolist()),
                                            properties={
                                                'time': f'{packet.time:%Y%m%d%H%M%S}',
                                                'callsign': packet.from_callsign,
                                                'altitude': packet.coordinates[2],
                                                'ascent_rate': packet_track.ascent_rates[packet_index],
                                                'ground_speed': packet_track.ground_speeds[packet_index]
                                            })
                            for packet_index, packet in enumerate(packet_track))

            features.append(geojson.Feature(geometry=geojson.LineString([packet.coordinates.tolist()
                                                                         for packet in packet_track.packets]),
                                            properties={
                                                'time': f'{packet_track.packets[-1].time:%Y%m%d%H%M%S}',
                                                'callsign': packet_track.callsign,
                                                'altitude': packet_track.coordinates[-1, -1],
                                                'ascent_rate': packet_track.ascent_rates[-1],
                                                'ground_speed': packet_track.ground_speeds[-1],
                                                'seconds_to_ground': packet_track.time_to_ground / timedelta(seconds=1)
                                            }))

        features = geojson.FeatureCollection(features)

        with open(output_filename, 'w') as output_file:
            geojson.dump(features, output_file)
    elif extension == '.kml':
        from fastkml import kml

        output_kml = kml.KML()
        document = kml.Document(KML_STANDARD, '1', 'root document', 'root document, containing geometries')
        output_kml.append(document)

        for packet_track_index, packet_track in enumerate(packet_tracks):
            for packet_index, packet in enumerate(packet_track):
                placemark = kml.Placemark(KML_STANDARD, f'1 {packet_track_index} {packet_index}',
                                          f'{packet_track.callsign} {packet.time:%Y%m%d%H%M%S}',
                                          f'altitude={packet.coordinates[2]} '
                                          f'ascent_rate={packet_track.ascent_rates[packet_index]} '
                                          f'ground_speed={packet_track.ground_speeds[packet_index]}')
                placemark.geometry = Point(packet.coordinates.tolist())
                document.append(placemark)

            placemark = kml.Placemark(KML_STANDARD, f'1 {packet_track_index}', packet_track.callsign,
                                      f'altitude={packet_track.coordinates[-1, -1]} '
                                      f'ascent_rate={packet_track.ascent_rates[-1]} '
                                      f'ground_speed={packet_track.ground_speeds[-1]} '
                                      f'seconds_to_ground={packet_track.time_to_ground / timedelta(seconds=1)}')
            placemark.geometry = LineString(packet_track.coordinates)
            document.append(placemark)

        with open(output_filename, 'w') as output_file:
            output_file.write(output_kml.to_string())
    else:
        raise NotImplementedError(f'saving to file type "{extension}" has not been implemented')



from datetime import timedelta
from typing import Union

import numpy
from pyproj import CRS, Geod

from .packets import APRSPacket, DEFAULT_CRS, LocationPacket
from .structures import DoublyLinkedList


class LocationPacketTrack:
    """ collection of location packets """

    def __init__(self, packets: [LocationPacket] = None, crs: CRS = None):
        """
        location packet track

        :param crs: coordinate reference system to use
        :param packets: iterable of packets
        """

        self.packets = DoublyLinkedList(packets)
        self.crs = crs if crs is not None else DEFAULT_CRS

    def append(self, packet: LocationPacket):
        if packet not in self.packets:
            if packet.crs != self.crs:
                packet.transform_to(self.crs)
            self.packets.append(packet)

    @property
    def times(self) -> numpy.ndarray:
        return numpy.array([packet.time for packet in self.packets], dtype=numpy.datetime64)

    @property
    def coordinates(self) -> numpy.ndarray:
        return numpy.stack([packet.coordinates for packet in self.packets], axis=0)

    @property
    def intervals(self) -> numpy.ndarray:
        return numpy.concatenate([[0], numpy.array([packet_delta.seconds for packet_delta in numpy.diff(self.packets)])])

    @property
    def distances(self) -> numpy.ndarray:
        return numpy.concatenate([[0], numpy.array([packet_delta.distance for packet_delta in numpy.diff(self.packets)])])

    @property
    def ascents(self) -> numpy.ndarray:
        return numpy.concatenate([[0], numpy.array([packet_delta.ascent for packet_delta in numpy.diff(self.packets)])])

    @property
    def ascent_rates(self) -> numpy.ndarray:
        return numpy.concatenate([[0], numpy.array([packet_delta.ascent_rate for packet_delta in numpy.diff(self.packets)])])

    @property
    def ground_speeds(self) -> numpy.ndarray:
        return numpy.concatenate([[0], numpy.array([packet_delta.ground_speed for packet_delta in numpy.diff(self.packets)])])

    @property
    def time_to_ground(self) -> timedelta:
        """ estimated time to reach the ground at the current ascent rate """

        current_ascent_rate = self.ascent_rates[-1]

        if current_ascent_rate < 0:
            # TODO implement landing location as the intersection of the predicted descent track with a local DEM
            # TODO implement a time to impact calc based off of standard atmo
            return timedelta(seconds=self.packets[-1].coordinates[2] / current_ascent_rate)
        else:
            return timedelta(seconds=-1)

    @property
    def distance_from_start(self) -> float:
        """ overground distance from the first to the most recent packet """

        if len(self.packets) > 0:
            return self.packets[-1].distance(self.coordinates[0, :2])
        else:
            return 0.0

    @property
    def length(self) -> float:
        """ total length of the packet track over the ground """

        coordinates = self.coordinates[:2]
        if self.crs.is_projected:
            return sum(self.packets.difference)
        else:
            ellipsoid = self.crs.datum.to_json_dict()['ellipsoid']
            geodetic = Geod(a=ellipsoid['semi_major_axis'], rf=ellipsoid['inverse_flattening'])
            return geodetic.line_length(coordinates[:, 0], coordinates[:, 1])

    def __getitem__(self, index: Union[int, slice]) -> LocationPacket:
        return self.packets[index]

    def __iter__(self):
        return iter(self.packets)

    def __reversed__(self):
        return reversed(self.packets)

    def __contains__(self, item) -> bool:
        return item in self.packets

    def __len__(self) -> int:
        return len(self.packets)

    def __eq__(self, other) -> bool:
        return self.packets == other.packets

    def __str__(self) -> str:
        return str(list(self))


class APRSTrack(LocationPacketTrack):
    """ collection of APRS location packets """

    def __init__(self, callsign: str, packets: [APRSPacket] = None, crs: CRS = None):
        """
        APRS packet track

        :param callsign: callsign of APRS packets
        :param packets: iterable of packets
        :param crs: coordinate reference system to use
        """

        self.callsign = callsign
        super().__init__(packets, crs)

    def append(self, packet: APRSPacket):
        packet_callsign = packet['callsign']

        if packet_callsign == self.callsign:
            super().append(packet)
        else:
            raise ValueError(f'Packet callsign {packet_callsign} does not match ground track callsign {self.callsign}.')

    def __eq__(self, other) -> bool:
        return self.callsign == other.__callsign and super().__eq__(other)

    def __str__(self) -> str:
        return f'{self.callsign}: {super().__str__()}'



from typing import Union

import aprslib


def parse_raw_aprs(raw_aprs: Union[str, dict]) -> dict:
    """
    Parse APRS fields from raw packet string.

    APRS format reference: http://www.aprs.org/doc/APRS101.PDF

    :param raw_aprs: raw APRS string
    :return: dictionary of APRS fields
    """

    if not isinstance(raw_aprs, dict):
        try:
            parsed_packet = aprslib.parse(raw_aprs)
        except aprslib.ParseError as error:
            raise InvalidPacketError(str(error))
    else:
        parsed_packet = {
            'from': raw_aprs['srccall'],
            'to': raw_aprs['dstcall'],
            'path': raw_aprs['path'].split(','),
            'timestamp': raw_aprs['time'],
            'symbol': raw_aprs['symbol'][1:],
            'symbol_table': raw_aprs['symbol'][0],
            'latitude': float(raw_aprs['lat']),
            'longitude': float(raw_aprs['lng']),
            'altitude': float(raw_aprs['altitude']) if 'altitude' in raw_aprs else None,
            'comment': raw_aprs['comment'] if 'comment' in raw_aprs else 'comment',
        }

    # parsed_packet = {'raw': str(raw_aprs)}
    #
    # if ':' not in raw_aprs:
    #     raise InvalidPacketError('location data missing')
    #
    # parsed_packet['from'], raw_aprs = raw_aprs.split('>', 1)
    # parsed_packet['to'], raw_aprs = raw_aprs.split(',', 1)
    # parsed_packet['path'], raw_aprs = raw_aprs.split(':', 1)
    # parsed_packet['path'] = parsed_packet['path'].split(',')
    # parsed_packet['via'] = parsed_packet['path'][-1]
    #
    # parsed_packet['messagecapable'] = raw_aprs[0] in ['!', '/']
    #
    # try:
    #     raw_aprs = raw_aprs[2:]
    #
    #     # TODO add parsing for uncompressed coordinates
    #     lonlat_string = raw_aprs[:8]
    #     parsed_packet['latitude'] = decompress_latitude(lonlat_string[:4])
    #     parsed_packet['longitude'] = decompress_longitude(lonlat_string[4:])
    #     parsed_packet['symbol'] = raw_aprs[8]
    #
    #     raw_aprs = raw_aprs[9:]
    #
    #     # TODO add parsing for more altitude formats
    #     # convert altitude from feet to meters
    #     parsed_packet['altitude'] = int(re.findall('/A=.{6}', raw_aprs)[0][3:]) * 0.3048
    #
    #     parsed_packet['comment'] = re.split('/A=.{6}', raw_aprs)[-1]
    # except IndexError:
    #     raise InvalidPacketError('packet terminated unexpectedly')

    return parsed_packet


class InvalidPacketError(Exception):
    pass


def decompress_longitude(compressed_longitude: str) -> float:
    """
    Decode longitude string from APRS compressed format (shifted ASCII in base 91) to a float.

    :param compressed_longitude: compressed APRS longitude string
    :return: longitude
    """

    converted_floats = []
    current_power = len(compressed_longitude) - 1

    for character in compressed_longitude:
        converted_floats.append(float(ord(character) - 33) * (91 ** current_power))
        current_power -= 1

    return -180 + (sum(converted_floats) / 190463)


def decompress_latitude(compressed_latitude: str) -> float:
    """
    Decode latitude string from APRS compressed format (shifted ASCII in base 91) to a float.

    :param compressed_latitude: compressed APRS latitude string
    :return: latitude
    """

    converted_floats = []
    current_power = len(compressed_latitude) - 1

    for character in compressed_latitude:
        converted_floats.append(float(ord(character) - 33) * (91 ** current_power))
        current_power -= 1

    return 90 - (sum(converted_floats) / 380926)



from argparse import ArgumentParser
from datetime import datetime
from getpass import getpass
import os
from pathlib import Path
import sys
import time

from dateutil.parser import parse as parse_date

from client import DEFAULT_INTERVAL_SECONDS
from client.gui import PacketRavenGUI
from client.retrieve import retrieve_packets
from packetraven.connections import APRSDatabaseTable, APRSfi, APRSis, SerialTNC, TextFileTNC
from packetraven.utilities import get_logger, read_configuration, repository_root

LOGGER = get_logger('packetraven')

CREDENTIALS_FILENAME = repository_root() / 'credentials.config'


def main():
    args_parser = ArgumentParser()
    args_parser.add_argument('--callsigns', help='comma-separated list of callsigns to track')
    args_parser.add_argument('--apikey', help='APRS.fi API key (from https://aprs.fi/page/api)')
    args_parser.add_argument('--tnc', help='serial port or text file of TNC parsing APRS packets from analog audio to ASCII'
                                           ' (set to `auto` to use the first open serial port)')
    args_parser.add_argument('--database', help='PostGres database table `user@hostname:port/database/table`')
    args_parser.add_argument('--tunnel', help='SSH tunnel `user@hostname:port`')
    args_parser.add_argument('--igate', action='store_true', help='send new packets to APRS-IS')
    args_parser.add_argument('--start', help='start date / time, in any common date format')
    args_parser.add_argument('--end', help='end date / time, in any common date format')
    args_parser.add_argument('--log', help='path to log file to save log messages')
    args_parser.add_argument('--output', help='path to output file to save packets')
    args_parser.add_argument('--interval', default=DEFAULT_INTERVAL_SECONDS, type=float,
                             help='seconds between each main loop (default: 5)')
    args_parser.add_argument('--gui', action='store_true', help='start the graphical interface')
    args = args_parser.parse_args()

    using_gui = args.gui
    using_igate = args.igate

    callsigns = [callsign.upper() for callsign in args.callsigns.strip('"').split(',')] if args.callsigns is not None else None

    kwargs = {}

    if args.apikey is not None:
        kwargs['api_key'] = args.apikey

    if args.tnc is not None:
        kwargs['tnc'] = args.tnc

    if args.database is not None:
        database = args.database
        if database.count('/') != 2:
            LOGGER.error(f'unable to parse hostname, database name, and table name from input "{database}"')
        else:
            kwargs['hostname'], kwargs['database'], kwargs['table'] = database.split('/')
            if '@' in kwargs['hostname']:
                kwargs['username'], kwargs['hostname'] = kwargs['hostname'].split('@', 1)
                if ':' in kwargs['username']:
                    kwargs['username'], kwargs['password'] = kwargs['username'].split(':', 1)

    if args.tunnel is not None:
        kwargs['ssh_hostname'] = args.tunnel
        if '@' in kwargs['ssh_hostname']:
            kwargs['ssh_username'], kwargs['ssh_hostname'] = kwargs['ssh_hostname'].split('@', 1)
            if ':' in kwargs['ssh_username']:
                kwargs['ssh_username'], kwargs['ssh_password'] = kwargs['ssh_username'].split(':', 1)

    if args.start is not None:
        kwargs['start_date'] = parse_date(args.start.strip('"'))

    if args.end is not None:
        kwargs['end_date'] = parse_date(args.end.strip('"'))

    if 'start_date' in kwargs and 'end_date' in kwargs:
        if kwargs['start_date'] > kwargs['end_date']:
            start_date = kwargs['start_date']
            kwargs['start_date'] = kwargs['end_date']
            kwargs['end_date'] = start_date
            del start_date

    if args.log is not None:
        log_filename = Path(args.log).expanduser()
        if not log_filename.exists():
            os.makedirs(log_filename, exist_ok=True)
        if log_filename.is_dir():
            log_filename = log_filename / f'packetraven_log_{datetime.now():%Y%m%dT%H%M%S}.txt'
        get_logger(LOGGER.name, log_filename)
    else:
        log_filename = None

    if args.output is not None:
        output_filename = Path(args.output).expanduser()
        if not output_filename.parent.exists():
            os.makedirs(output_filename.parent, exist_ok=True)
        if output_filename.is_dir():
            output_filename = output_filename / f'packetraven_output_{datetime.now():%Y%m%dT%H%M%S}.geojson'
    else:
        output_filename = None

    interval_seconds = args.interval if args.interval >= 1 else 1

    if CREDENTIALS_FILENAME.exists():
        credentials = {}
        for section in read_configuration(CREDENTIALS_FILENAME).values():
            credentials.update(section)
        kwargs = {key: value if key not in kwargs or kwargs[key] is None else kwargs[key]
                  for key, value in credentials.items()}

    if using_gui:
        PacketRavenGUI(callsigns, log_filename, output_filename, interval_seconds, using_igate, **kwargs)
    else:
        start_date = kwargs['start_date'] if 'start_date' in kwargs else None
        end_date = kwargs['end_date'] if 'end_date' in kwargs else None

        connections = []
        if 'tnc' in kwargs:
            tnc_kwargs = {key: kwargs[key] for key in ['tnc'] if key in kwargs}
            if tnc_kwargs['tnc'] is not None and 'txt' in tnc_kwargs['tnc']:
                try:
                    text_file_tnc = TextFileTNC(kwargs['tnc'], callsigns)
                    LOGGER.info(f'reading file {text_file_tnc.location}')
                    connections.append(text_file_tnc)
                except ConnectionError as error:
                    LOGGER.warning(f'{error.__class__.__name__} - {error}')
            else:
                try:
                    serial_tnc = SerialTNC(kwargs['tnc'], callsigns)
                    LOGGER.info(f'opened port {serial_tnc.location}')
                    connections.append(serial_tnc)
                except ConnectionError as error:
                    LOGGER.warning(f'{error.__class__.__name__} - {error}')

        if 'api_key' in kwargs:
            aprs_fi_kwargs = {key: kwargs[key] for key in ['api_key'] if key in kwargs}
            try:
                aprs_api = APRSfi(callsigns=callsigns, **aprs_fi_kwargs)
                LOGGER.info(f'connected to {aprs_api.location}')
                connections.append(aprs_api)
            except ConnectionError as error:
                LOGGER.warning(f'{error.__class__.__name__} - {error}')

        if 'hostname' in kwargs:
            database_kwargs = {key: kwargs[key] for key in ['hostname', 'database', 'table', 'username', 'password', ]
                               if key in kwargs}
            ssh_tunnel_kwargs = {key: kwargs[key] for key in ['ssh_hostname', 'ssh_username', 'ssh_password']
                                 if key in kwargs}

            try:
                if 'ssh_hostname' in ssh_tunnel_kwargs:
                    if 'ssh_username' not in ssh_tunnel_kwargs or ssh_tunnel_kwargs['ssh_username'] is None:
                        ssh_username = input(f'enter username for SSH host "{ssh_tunnel_kwargs["ssh_hostname"]}": ')
                        if ssh_username is None or len(ssh_username) == 0:
                            raise ConnectionError('missing SSH username')
                        ssh_tunnel_kwargs['ssh_username'] = ssh_username

                    if 'ssh_password' not in ssh_tunnel_kwargs or ssh_tunnel_kwargs['ssh_password'] is None:
                        ssh_password = getpass(f'enter password for SSH user "{ssh_tunnel_kwargs["ssh_username"]}": ')
                        if ssh_password is None or len(ssh_password) == 0:
                            raise ConnectionError('missing SSH password')
                        ssh_tunnel_kwargs['ssh_password'] = ssh_password

                if 'username' not in database_kwargs or database_kwargs['username'] is None:
                    database_username = input(f'enter username for database '
                                              f'"{database_kwargs["hostname"]}/{database_kwargs["database"]}": ')
                    if database_username is None or len(database_username) == 0:
                        raise ConnectionError('missing database username')
                    database_kwargs['username'] = database_username

                if 'password' not in database_kwargs or database_kwargs['password'] is None:
                    database_password = getpass(f'enter password for database user "{database_kwargs["username"]}": ')
                    if database_password is None or len(database_password) == 0:
                        raise ConnectionError('missing database password')
                    database_kwargs['password'] = database_password

                database = APRSDatabaseTable(**database_kwargs, **ssh_tunnel_kwargs, callsigns=callsigns)
                LOGGER.info(f'connected to {database.location}')
                connections.append(database)
            except ConnectionError:
                database = None
        else:
            database = None

        if using_igate:
            try:
                aprs_is = APRSis(callsigns)
            except ConnectionError:
                aprs_is = None
        else:
            aprs_is = None

        if len(connections) == 0:
            LOGGER.error(f'no connections started')
            sys.exit(1)

        filter_message = 'retrieving packets'
        if start_date is not None and end_date is None:
            filter_message += f' sent after {start_date:%Y-%m-%d %H:%M:%S}'
        elif start_date is None and end_date is not None:
            filter_message += f' sent before {end_date:%Y-%m-%d %H:%M:%S}'
        elif start_date is not None and end_date is not None:
            filter_message += f' sent between {start_date:%Y-%m-%d %H:%M:%S} and {end_date:%Y-%m-%d %H:%M:%S}'
        if callsigns is not None:
            filter_message += f' from {len(callsigns)} callsigns: {callsigns}'
        LOGGER.info(filter_message)

        LOGGER.info(f'listening for packets every {interval_seconds}s from {len(connections)} connection(s): '
                    f'{", ".join([connection.location for connection in connections])}')

        packet_tracks = {}
        try:
            while len(connections) > 0:
                parsed_packets = retrieve_packets(connections, packet_tracks, database, output_filename, start_date=start_date, end_date=end_date,
                                                  logger=LOGGER)
                if aprs_is is not None:
                    aprs_is.send(parsed_packets)
                time.sleep(interval_seconds)
        except KeyboardInterrupt:
            for connection in connections:
                connection.close()
            sys.exit(0)


if __name__ == '__main__':
    main()
